

from fastapi import APIRouter
from fastapi import Depends
from fastapi import BackgroundTasks

from tendril.authn.users import auth_spec
from tendril.authn.users import authn_dependency
from tendril.authn.users import AuthUserModel

from tendril.config import INTERESTS_API_ENABLED
from tendril.iotedge.profiles import device_config_unified_model

from tendril.iotcore.settings import get_device_settings
from tendril.iotcore.settings import set_device_settings
from tendril.iotcore.logs import request_device_logs
from tendril.iotcore.logs import available_device_logs

from tendril.utils.db import get_session
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


device_management_router = APIRouter(
    prefix='/devices', tags=["IoT Device Hub Management API"],
    dependencies=[Depends(authn_dependency)])


@device_management_router.get("/{id}/settings",
                              response_model=device_config_unified_model)
async def read_device_settings(id: int,
                               user: AuthUserModel = auth_spec(scopes=['device:read'])):
    with get_session() as session:
        return get_device_settings(id=id, auth_user=user,
                                   session=session).export(expand=False)


@device_management_router.post("/{id}/settings",
                               response_model=device_config_unified_model)
async def write_device_settings(id: int, settings: device_config_unified_model,
                                user: AuthUserModel = auth_spec(scopes=['device:write']),
                                background_tasks=BackgroundTasks):
    with get_session() as session:
        # TODO Clean this up.
        #  We are most probably never using these responses.
        #  Confirm and remove.
        result = set_device_settings(id=id, settings=settings,
                                     background_tasks=background_tasks,
                                     auth_user=user, session=session)
        return get_device_settings(id=id, auth_user=user,
                                   session=session).export(expand=False)


@device_management_router.post("/{id}/logs")
async def request_logs_from_device(id: int, background_tasks: BackgroundTasks,
                                   user: AuthUserModel = auth_spec(scopes=['device:write'])):
    with get_session() as session:
        return request_device_logs(id=id, background_tasks=background_tasks,
                                   auth_user=user, session=session)


@device_management_router.get("/{id}/logs")
async def get_available_logs(id: int, background_tasks: BackgroundTasks,
                             user: AuthUserModel = auth_spec(scopes=['device:write'])):
    with get_session() as session:
        return available_device_logs(id=id, background_tasks=background_tasks,
                                     auth_user=user, session=session)


if INTERESTS_API_ENABLED:
    routers = [
        device_management_router
    ]
else:
    logger.info("Not creating Device Management API routers.")
    routers = []





from tendril.core.mq.aio import with_mq_client
from tendril.utils import log
logger = log.get_logger(__name__)


@with_mq_client('X')
async def create_mq_topology(mq=None):
    logger.info("Creating the Device Communications queue.")
    device_aggregation_queue = await mq.create_work_queue('device_aggregation', topic='device.from.*')





from functools import wraps
from tendril import interests
from tendril.db.controllers.interests import get_interest

from tendril.common.iotedge.exceptions import DeviceTypeUnrecognized
from tendril.common.iotedge.exceptions import DeviceTypeMismatch

from . import profiles

from tendril.utils.db import with_db
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


@with_db
def get_registration(device_id, appname=None, session=None):
    if appname:
        profile = profiles.profile(appname, device_id=device_id)
        interest = get_interest(name=device_id, type=profile.interest_type,
                                raise_if_none=False, session=session)
    else:
        interest = get_interest(name=device_id, raise_if_none=False,
                                session=session)
        appname = interest.appname
        profile = profiles.profile(appname, device_id=device_id)

    if not interest:
        return None

    if appname and interest.appname != appname:
        raise DeviceTypeMismatch(appname, interest.appname, device_id)

    return profile(interest)


@with_db
def register(device_id, appname, info={}, session=None):
    profile = profiles.profile(appname, device_id=device_id)
    itype = interests.type_codes[profile.interest_type]
    interest = itype(name=device_id, appname=appname, info=info,
                     must_create=True, session=session)
    return profile(interest)


def registered_device(appname=None):
    def decorator(func):
        @wraps(func)
        def get_registered_device(device_id=None, session=None, **kwargs):
            if appname:
                device = get_registration(device_id, appname=appname, session=session)
            else:
                device = get_registration(device_id, session=session)
                kwargs['appname'] = device.appname
            return func(device=device, session=session, **kwargs)
        return get_registered_device
    return decorator





from tendril.libraries import interests
from tendril.common.interests.exceptions import InterestTypeUnsupported

from tendril.utils.db import with_db
from tendril.utils import log
logger = log.get_logger(__name__)


@with_db
def get_device_settings(id: int, auth_user=None, session=None):
    library = interests.find_library(id)
    interest = library.item(id=id, session=session)
    if not hasattr(interest, 'appname'):
        raise InterestTypeUnsupported("'appname' attribute", id=id, name=interest.name)
    profile = interest.profile(auth_user=auth_user, session=session)
    return profile.config(session=session)


@with_db
def set_device_settings(id: int, settings=None, background_tasks=None, auth_user=None, session=None):
    library = interests.find_library(id)
    interest = library.item(id=id, session=session)
    if not hasattr(interest, 'appname'):
        raise InterestTypeUnsupported("'appname' attribute", id=id, name=interest.name)
    response = interest.configure(settings=settings, auth_user=auth_user, session=session)
    interest.trigger_device_settings_update(background_tasks=background_tasks)
    return response





from typing import Optional
from tendril.utils.pydantic import TendrilTBaseModel
from tendril.common.states import LifecycleStatus


IoTDeviceIDTModel = str
IoTDeviceAppnameTModel = str


class IoTDeviceMessageTModel(TendrilTBaseModel):
    id: IoTDeviceIDTModel


class IoTDeviceAnnounceTModel(IoTDeviceMessageTModel):
    appname: IoTDeviceAppnameTModel
    have_credentials: bool
    sysinfo: Optional[dict]


class IoTDevicePingTModel(IoTDeviceMessageTModel):
    status: Optional[dict]


class IoTDeviceAnnounceResponseTModel(IoTDeviceMessageTModel):
    status: LifecycleStatus
    interest_id: int
    password: Optional[str]


class IoTDeviceSettingRequestTModel(IoTDeviceMessageTModel):
    appname: IoTDeviceAppnameTModel





from tendril.common.exceptions import HTTPCodedException


class DeviceTypeUnrecognized(HTTPCodedException):
    status_code = 406

    def __init__(self, appname, id=None):
        self.appname = appname
        self.id = id

    def __str__(self):
        if not self.id:
            return f"Device type '{self.appname}' is unrecognized."
        return f"Device type provided by '{self.id}', '{self.appname}' is unrecognized."


class DeviceTypeMismatch(HTTPCodedException):
    status_code = 409

    def __init__(self, actual, expected, id=None):
        self.actual = actual
        self.expected = expected
        self.id = id

    def __str__(self):
        if not self.id:
            return f"Device type mismatch. Got '{self.actual}', expecting '{self.expected}'"
        return f"Device type of '{self.id}' ('{self.actual}') " \
               f"does not match the required or registered type " \
               f"'{self.expected}'"
